@@ -4,7 +4,7 @@
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 module Data.Aeson.Config.Types where
 
-import           Data.Monoid hiding (Product)
+import           Data.Semigroup (Semigroup(..))
 import           Data.Bitraversable
 import           Data.Bifoldable
 import           Data.Bifunctor
@@ -12,7 +12,7 @@ import           Data.Bifunctor
 import           Data.Aeson.Config.FromValue
 
 newtype List a = List {fromList :: [a]}
-  deriving (Eq, Show, Functor, Foldable, Traversable, Monoid)
+  deriving (Eq, Show, Functor, Foldable, Traversable, Semigroup, Monoid)
 
 instance FromValue a => FromValue (List a) where
   fromValue v = List <$> case v of
@@ -25,9 +25,12 @@ fromMaybeList = maybe [] fromList
 data Product a b = Product a b
   deriving (Eq, Show, Functor, Foldable, Traversable)
 
-instance (Monoid a, Monoid b) => Monoid (Product a b) where
+instance (Semigroup a, Semigroup b, Monoid a, Monoid b) => Monoid (Product a b) where
   mempty = Product mempty mempty
-  Product a1 b1 `mappend` Product a2 b2 = Product (a1 <> a2) (b1 <> b2)
+  mappend = (<>)
+
+instance (Semigroup a, Semigroup b) => Semigroup (Product a b) where
+  Product a1 b1 <> Product a2 b2 = Product (a1 <> a2) (b1 <> b2)
 
 instance Bifunctor Product where
   bimap fa fb (Product a b) = Product (fa a) (fb b)
