@@ -16,11 +16,14 @@ import           Data.Maybe
 import           Data.List
 import           Data.String.Interpolate
 import           Data.String.Interpolate.Util
+import           Data.Version (showVersion)
 
 import qualified Hpack.Render as Hpack
 import           Hpack.Config (packageConfig, readPackageConfig, DecodeOptions(..), DecodeResult(..), defaultDecodeOptions)
 import           Hpack.Render.Hints (FormattingHints(..), sniffFormattingHints)
 
+import qualified Paths_hpack as Hpack (version)
+
 writeFile :: FilePath -> String -> IO ()
 writeFile file c = touch file >> Prelude.writeFile file c
 
@@ -37,6 +40,46 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
           Paths_foo
       |]
 
+    it "warns on duplicate fields" $ do
+      [i|
+      name: foo
+      name: foo
+      |] `shouldWarn` [
+          "package.yaml: Duplicate field $.name"
+        ]
+
+    describe "spec-version" $ do
+      it "accepts spec-version" $ do
+        [i|
+        spec-version: 0.29.5
+        |] `shouldRenderTo` package [i|
+        |]
+
+      it "fails on malformed spec-version" $ do
+        [i|
+        spec-version: foo
+        |] `shouldFailWith` "package.yaml: Error while parsing $.spec-version - invalid value \"foo\""
+
+      it "fails on unsupported spec-version" $ do
+        [i|
+        spec-version: 25.0
+        dependencies: foo == bar
+        |] `shouldFailWith` ("The file package.yaml requires version 25.0 of the Hpack package specification, however this version of hpack only supports versions up to " ++ showVersion Hpack.version ++ ". Upgrading to the latest version of hpack may resolve this issue.")
+
+      it "fails on unsupported spec-version from defaults" $ do
+        let file = joinPath ["defaults", "sol", "hpack-template", "2017", "defaults.yaml"]
+        writeFile file [i|
+        spec-version: 25.0
+        |]
+
+        [i|
+        defaults:
+          github: sol/hpack-template
+          path: defaults.yaml
+          ref: "2017"
+        library: {}
+        |] `shouldFailWith` ("The file " ++ file ++ " requires version 25.0 of the Hpack package specification, however this version of hpack only supports versions up to " ++ showVersion Hpack.version ++ ". Upgrading to the latest version of hpack may resolve this issue.")
+
     describe "data-files" $ do
       it "accepts data-files" $ do
         touch "data/foo/index.html"
@@ -299,6 +342,82 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         version: {}
         |] `shouldFailWith` "package.yaml: Error while parsing $.version - expected Number or String, encountered Object"
 
+    describe "license" $ do
+      it "accepts cabal-style licenses" $ do
+        [i|
+        license: BSD3
+        |] `shouldRenderTo` (package [i|
+        license: BSD3
+        |])
+
+      it "accepts SPDX licenses" $ do
+        [i|
+        license: BSD-3-Clause
+        |] `shouldRenderTo` (package [i|
+        license: BSD-3-Clause
+        |]) {packageCabalVersion = "2.2"}
+
+      context "with an ambiguous license" $ do
+        it "treats it as a cabal-style license" $ do
+          [i|
+          license: MIT
+          |] `shouldRenderTo` (package [i|
+          license: MIT
+          |])
+
+      context "when cabal-version >= 2.2" $ do
+        it "maps license to SPDX license identifier" $ do
+          [i|
+          license: BSD3
+          library:
+            cxx-options: -Wall
+          |] `shouldRenderTo` (package [i|
+          license: BSD-3-Clause
+          library
+            other-modules:
+                Paths_foo
+            cxx-options: -Wall
+            default-language: Haskell2010
+          |]) {packageCabalVersion = "2.2"}
+
+        it "doesn't touch unknown licenses" $ do
+          [i|
+          license: some-license
+          library:
+            cxx-options: -Wall
+          |] `shouldRenderTo` (package [i|
+          license: some-license
+          library
+            other-modules:
+                Paths_foo
+            cxx-options: -Wall
+            default-language: Haskell2010
+          |]) {packageCabalVersion = "2.2"}
+
+      context "with a LICENSE file" $ do
+        before_ (writeFile "LICENSE" license) $ do
+          it "infers license" $ do
+            [i|
+            |] `shouldRenderTo` (package [i|
+            license-file: LICENSE
+            license: MIT
+            |])
+
+          context "when license can not be inferred" $ do
+            it "warns" $ do
+              writeFile "LICENSE" "some-licenese"
+              [i|
+              name: foo
+              |] `shouldWarn` ["Inferring license from file LICENSE failed!"]
+
+          context "when license is null" $ do
+            it "does not infer license" $ do
+              [i|
+              license: null
+              |] `shouldRenderTo` (package [i|
+              license-file: LICENSE
+              |])
+
     describe "build-type" $ do
       it "accept Simple" $ do
         [i|
@@ -338,7 +457,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         extra-doc-files:
             CHANGES.markdown
             README.markdown
-        |]) {packageCabalVersion = ">= 1.18"}
+        |]) {packageCabalVersion = "1.18"}
 
       it "accepts glob patterns" $ do
         touch "CHANGES.markdown"
@@ -350,7 +469,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         extra-doc-files:
             CHANGES.markdown
             README.markdown
-        |]) {packageCabalVersion = ">= 1.18"}
+        |]) {packageCabalVersion = "1.18"}
 
       it "warns if a glob pattern does not match anything" $ do
         [i|
@@ -359,6 +478,124 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
           - "*.markdown"
         |] `shouldWarn` ["Specified pattern \"*.markdown\" for extra-doc-files does not match any files"]
 
+    describe "build-tools" $ do
+      it "adds known build tools to build-tools" $ do
+        [i|
+        executable:
+          build-tools:
+            alex == 0.1.0
+        |] `shouldRenderTo` executable_ "foo" [i|
+        build-tools:
+            alex ==0.1.0
+        |]
+
+      it "adds other build tools to build-tool-depends" $ do
+        [i|
+        executable:
+          build-tools:
+            hspec-discover: 0.1.0
+        |] `shouldRenderTo` (executable_ "foo" [i|
+        build-tool-depends:
+            hspec-discover:hspec-discover ==0.1.0
+        |]) {
+          -- NOTE: We do not set this to 2.0 on purpose, so that the .cabal
+          -- file is compatible with a wider range of Cabal versions!
+          packageCabalVersion = "1.12"
+        }
+
+      context "when the name of a build tool matches an executable from the same package" $ do
+        it "adds it to build-tools" $ do
+          [i|
+          executables:
+            bar:
+              build-tools:
+                - bar
+          |] `shouldRenderTo` executable_ "bar" [i|
+          build-tools:
+              bar
+          |]
+
+        it "gives per-section unqualified names precedence over global qualified names" $ do
+          [i|
+          build-tools:
+            - foo:bar == 0.1.0
+          executables:
+            bar:
+              build-tools:
+                - bar == 0.2.0
+          |] `shouldRenderTo` executable_ "bar" [i|
+          build-tools:
+              bar ==0.2.0
+          |]
+
+        it "gives per-section qualified names precedence over global unqualified names" $ do
+          [i|
+          build-tools:
+            - bar == 0.1.0
+          executables:
+            bar:
+              build-tools:
+                - foo:bar == 0.2.0
+          |] `shouldRenderTo` executable_ "bar" [i|
+          build-tools:
+              bar ==0.2.0
+          |]
+
+      context "when the name of a build tool matches a legacy system build tool" $ do
+        it "adds it to build-tools" $ do
+          [i|
+          executable:
+            build-tools:
+              ghc >= 7.10
+          |] `shouldRenderTo` (executable_ "foo" [i|
+          build-tools:
+              ghc >=7.10
+          |]) { packageWarnings = ["Listing \"ghc\" under build-tools is deperecated! Please list system executables under system-build-tools instead!"] }
+
+    describe "system-build-tools" $ do
+      it "adds system build tools to build-tools" $ do
+        [i|
+        executable:
+          system-build-tools:
+            ghc >= 7.10
+        |] `shouldRenderTo` executable_ "foo" [i|
+        build-tools:
+            ghc >=7.10
+        |]
+
+      context "with hpc" $ do
+        it "infers cabal-version 1.14" $ do
+          [i|
+          executable:
+            system-build-tools:
+              hpc
+          |] `shouldRenderTo` (executable_ "foo" [i|
+          build-tools:
+              hpc
+          |]) {packageCabalVersion = "1.14"}
+
+      context "with ghcjs" $ do
+        it "infers cabal-version 1.22" $ do
+          [i|
+          executable:
+            system-build-tools:
+              ghcjs
+          |] `shouldRenderTo` (executable_ "foo" [i|
+          build-tools:
+              ghcjs
+          |]) {packageCabalVersion = "1.22"}
+
+      context "with an unknown system build tool" $ do
+        it "infers cabal-version 2.0" $ do
+          [i|
+          executable:
+            system-build-tools:
+              g++ >= 5.4.0
+          |] `shouldRenderTo` (executable_ "foo" [i|
+          build-tools:
+              g++ >=5.4.0
+          |]) {packageCabalVersion = "2.0"}
+
     describe "dependencies" $ do
       it "accepts single dependency" $ do
         [i|
@@ -476,7 +713,25 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
           cxx-options: -Wall
         |] `shouldRenderTo` (executable_ "foo" [i|
         cxx-options: -Wall
-        |]) {packageCabalVersion = ">= 2.2"}
+        |]) {packageCabalVersion = "2.2"}
+
+      context "when used inside a nested conditional" $ do
+        it "infers correct cabal-version" $ do
+          [i|
+          executable:
+            when:
+              condition: True
+              when:
+                condition: True
+                when:
+                  condition: True
+                  cxx-options: -Wall
+          |] `shouldRenderTo` (executable_ "foo" [i|
+          if true
+            if true
+              if true
+                cxx-options: -Wall
+          |]) {packageCabalVersion = "2.2"}
 
     describe "cxx-sources" $ before_ (touch "foo.cc" >> touch "cxxbits/bar.cc") $ do
       it "accepts cxx-sources" $ do
@@ -487,9 +742,9 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
             - cxxbits/*.cc
         |] `shouldRenderTo` (executable_ "foo" [i|
         cxx-sources:
-            cxxbits/bar.cc
             foo.cc
-        |]) {packageCabalVersion = ">= 2.2"}
+            cxxbits/bar.cc
+        |]) {packageCabalVersion = "2.2"}
 
     describe "extra-lib-dirs" $ do
       it "accepts extra-lib-dirs" $ do
@@ -543,72 +798,50 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
             bar
         |]
 
-    describe "c-sources" $ before_ (touch "cbits/foo.c" >> touch "cbits/bar.c") $ do
-      context "with internal-libraries" $ do
-        it "warns when a glob pattern does not match any files" $ do
-          [i|
-          name: foo
-          internal-libraries:
-            bar:
-              c-sources: foo/*.c
-          |] `shouldWarn` pure "Specified pattern \"foo/*.c\" for c-sources does not match any files"
+    describe "c-sources" $ before_ (touch "cbits/foo.c" >> touch "cbits/bar.c" >> touch "cbits/baz.c") $ do
+      it "keeps declaration order" $ do
+        -- IMPORTANT: This is crucial as a workaround for https://ghc.haskell.org/trac/ghc/ticket/13786
+        [i|
+        library:
+          c-sources:
+            - cbits/foo.c
+            - cbits/bar.c
+            - cbits/baz.c
+        |] `shouldRenderTo` library_ [i|
+        c-sources:
+            cbits/foo.c
+            cbits/bar.c
+            cbits/baz.c
+        |]
 
-      context "with library" $ do
-        it "accepts global c-sources" $ do
-          [i|
+      it "accepts glob patterns" $ do
+        [i|
+        library:
           c-sources: cbits/*.c
-          library: {}
-          |] `shouldRenderTo` library_ [i|
-          c-sources:
-              cbits/bar.c
-              cbits/foo.c
-          |]
+        |] `shouldRenderTo` library_ [i|
+        c-sources:
+            cbits/bar.c
+            cbits/baz.c
+            cbits/foo.c
+        |]
 
-        it "accepts c-sources" $ do
-          [i|
-          library:
-            c-sources: cbits/*.c
-          |] `shouldRenderTo` library_ [i|
-          c-sources:
-              cbits/bar.c
-              cbits/foo.c
-          |]
+      it "warns when a glob pattern does not match any files" $ do
+        [i|
+        name: foo
+        library:
+          c-sources: foo/*.c
+        |] `shouldWarn` pure "Specified pattern \"foo/*.c\" for c-sources does not match any files"
 
-        it "accepts c-sources in conditional" $ do
-          [i|
-          library:
-            when:
-              condition: os(windows)
-              c-sources: cbits/*.c
-          |] `shouldRenderTo` library_ [i|
-          if os(windows)
-            c-sources:
-                cbits/bar.c
-                cbits/foo.c
-          |]
-
-      context "with executables" $ do
-        it "accepts global c-sources" $ do
-          [i|
-          c-sources: cbits/*.c
-          executables:
-            foo: {}
-          |] `shouldRenderTo` executable_ "foo" [i|
+      it "quotes filenames with special characters" $ do
+        touch "cbits/foo bar.c"
+        [i|
+        library:
           c-sources:
-              cbits/bar.c
-              cbits/foo.c
-          |]
-
-        it "accepts c-sources" $ do
-          [i|
-          executables:
-            foo:
-              c-sources: cbits/*.c
-          |] `shouldRenderTo` executable_ "foo" [i|
-          c-sources:
-              cbits/bar.c
-              cbits/foo.c
-          |]
+            - cbits/foo bar.c
+        |] `shouldRenderTo` library_ [i|
+        c-sources:
+            "cbits/foo bar.c"
+        |]
 
     describe "custom-setup" $ do
       it "warns on unknown fields" $ do
@@ -651,7 +884,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         |] `shouldRenderTo` (library_ [i|
           reexported-modules:
               Baz
-        |]) {packageCabalVersion = ">= 1.22"}
+        |]) {packageCabalVersion = "1.22"}
 
       it "accepts signatures" $ do
         [i|
@@ -660,7 +893,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         |] `shouldRenderTo` (library_ [i|
           signatures:
               Foo
-        |]) {packageCabalVersion = ">= 2.0"}
+        |]) {packageCabalVersion = "2.0"}
 
       context "when package.yaml contains duplicate modules" $ do
         it "generates a cabal file with duplicate modules" $ do
@@ -829,7 +1062,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
             autogen-modules:
                 Foo
                 Bar
-            |]) {packageCabalVersion = ">= 2.0"}
+            |]) {packageCabalVersion = "2.0"}
 
           it "does not infer any mentioned generated modules" $ do
             touch "src/Exposed.hs"
@@ -850,7 +1083,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
             autogen-modules:
                 Exposed
                 Other
-            |]) {packageCabalVersion = ">= 2.0"}
+            |]) {packageCabalVersion = "2.0"}
 
           it "does not infer any generated modules mentioned inside conditionals" $ do
             touch "src/Exposed.hs"
@@ -875,7 +1108,24 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
               autogen-modules:
                   Other
                   Exposed
-            |]) {packageCabalVersion = ">= 2.0"}
+            |]) {packageCabalVersion = "2.0"}
+
+      context "mixins" $ do
+        it "sets cabal-version to 2.0 if mixins are used" $ do
+          [i|
+          library:
+            dependencies:
+              foo:
+                mixin:
+                  - (Blah as Etc)
+          |] `shouldRenderTo` (library [i|
+          other-modules:
+              Paths_foo
+          build-depends:
+              foo
+          mixins:
+              foo (Blah as Etc)
+          |]) {packageCabalVersion = "2.0"}
 
     describe "internal-libraries" $ do
       it "accepts internal-libraries" $ do
@@ -980,7 +1230,7 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
                 Foo
             autogen-modules:
                 Foo
-          |]) {packageCabalVersion = ">= 2.0"}
+          |]) {packageCabalVersion = "2.0"}
 
         context "with conditional" $ do
           it "doesn't infer any modules mentioned in that conditional" $ do
@@ -1192,8 +1442,8 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
         it "overrides header fields" $ do
           [i|
           verbatim:
-            cabal-version: ">= 2.0"
-          |] `shouldRenderTo` (package "") {packageCabalVersion = ">= 2.0"}
+            cabal-version: foo
+          |] `shouldRenderTo` (package "") {packageCabalVersion = "foo"}
 
         it "overrides other fields" $ do
           touch "foo"
@@ -1232,6 +1482,30 @@ spec = around_ (inTempDirectoryNamed "foo") $ do
                 Paths_foo
             default-language: Haskell2010
           |]
+    describe "default value of maintainer" $ do
+      it "gives maintainer precedence" $ do
+        [i|
+          author: John Doe
+          maintainer: Jane Doe
+          |] `shouldRenderTo` package [i|
+          author: John Doe
+          maintainer: Jane Doe
+          |]
+      context "with author" $ do
+        it "uses author if maintainer is not specified" $ do
+          [i|
+            author: John Doe
+            |] `shouldRenderTo` package [i|
+            author: John Doe
+            maintainer: John Doe
+            |]
+        it "omits maintainer if it is null" $ do
+          [i|
+            author: John Doe
+            maintainer: null
+            |] `shouldRenderTo` package [i|
+            author: John Doe
+            |]
 
 run :: HasCallStack => FilePath -> FilePath -> String -> IO ([String], String)
 run userDataDir c old = run_ userDataDir c old >>= either assertFailure return
@@ -1240,12 +1514,12 @@ run_ :: FilePath -> FilePath -> String -> IO (Either String ([String], String))
 run_ userDataDir c old = do
   mPackage <- readPackageConfig defaultDecodeOptions {decodeOptionsTarget = c, decodeOptionsUserDataDir = Just userDataDir}
   return $ case mPackage of
-    Right (DecodeResult pkg _ warnings) ->
+    Right (DecodeResult pkg cabalVersion _ warnings) ->
       let
         FormattingHints{..} = sniffFormattingHints (lines old)
         alignment = fromMaybe 0 formattingHintsAlignment
         settings = formattingHintsRenderSettings
-        output = Hpack.renderPackageWith settings alignment formattingHintsFieldOrder formattingHintsSectionsFieldOrder pkg
+        output = cabalVersion ++ Hpack.renderPackageWith settings alignment formattingHintsFieldOrder formattingHintsSectionsFieldOrder pkg
       in
         Right (warnings, output)
     Left err -> Left err
@@ -1254,7 +1528,7 @@ data RenderResult = RenderResult [String] String
   deriving Eq
 
 instance Show RenderResult where
-  show (RenderResult warnings output) = unlines (map ("WARNING in " ++) warnings) ++ output
+  show (RenderResult warnings output) = unlines (map ("WARNING: " ++) warnings) ++ output
 
 shouldRenderTo :: HasCallStack => String -> Package -> Expectation
 shouldRenderTo input p = do
@@ -1263,7 +1537,7 @@ shouldRenderTo input p = do
   createDirectory currentDirectory
   withCurrentDirectory currentDirectory $ do
     (warnings, output) <- run ".." (".." </> packageConfig) expected
-    RenderResult warnings (dropEmptyLines output) `shouldBe` RenderResult [] expected
+    RenderResult warnings (dropEmptyLines output) `shouldBe` RenderResult (packageWarnings p) expected
   where
     expected = dropEmptyLines (renderPackage p)
     dropEmptyLines = unlines . filter (not . null) . lines
@@ -1280,7 +1554,7 @@ shouldFailWith input expected = do
   run_ "" packageConfig "" `shouldReturn` Left expected
 
 customSetup :: String -> Package
-customSetup a = (package content) {packageCabalVersion = ">= 1.24", packageBuildType = "Custom"}
+customSetup a = (package content) {packageCabalVersion = "1.24", packageBuildType = "Custom"}
   where
     content = [i|
 custom-setup
@@ -1308,7 +1582,7 @@ library
 |]
 
 internalLibrary :: String -> String -> Package
-internalLibrary name e = (package content) {packageCabalVersion = ">= 2.0"}
+internalLibrary name e = (package content) {packageCabalVersion = "2.0"}
   where
     content = [i|
 library #{name}
@@ -1322,7 +1596,7 @@ executable_ name e = package content
     content = [i|
 executable #{name}
   other-modules:
-      Paths_#{name}
+      Paths_foo
 #{indentBy 2 $ unindent e}
   default-language: Haskell2010
 |]
@@ -1337,7 +1611,7 @@ executable #{name}
 |]
 
 package :: String -> Package
-package = Package "foo" "0.0.0" "Simple" ">= 1.10"
+package c = Package "foo" "0.0.0" "Simple" "1.12" c []
 
 data Package = Package {
   packageName :: String
@@ -1345,17 +1619,41 @@ data Package = Package {
 , packageBuildType :: String
 , packageCabalVersion :: String
 , packageContent :: String
+, packageWarnings :: [String]
 }
 
 renderPackage :: Package -> String
 renderPackage Package{..} = unindent [i|
+cabal-version: #{packageCabalVersion}
 name: #{packageName}
 version: #{packageVersion}
 build-type: #{packageBuildType}
-cabal-version: #{packageCabalVersion}
 
 #{unindent packageContent}
 |]
 
 indentBy :: Int -> String -> String
 indentBy n = unlines . map (replicate n ' ' ++) . lines
+
+license :: String
+license = [i|
+Copyright (c) 2014-2018 Simon Hengel <sol@typeful.net>
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+|]
