@@ -12,6 +12,7 @@ module Hpack.ConfigSpec (
 
 , package
 , deps
+, defaultInfo
 ) where
 
 import           Helper
@@ -23,8 +24,11 @@ import qualified GHC.Exts as Exts
 import           System.Directory (createDirectory)
 import           Data.Either
 import qualified Data.Map.Lazy as Map
+import           Control.Monad.Trans.Writer (runWriter)
 
-import           Hpack.Syntax.Dependency
+import           Hpack.Syntax.Dependencies
+import           Hpack.Syntax.DependencyVersion
+import           Hpack.Syntax.BuildTools
 import           Hpack.Config hiding (package)
 import qualified Hpack.Config as Config
 
@@ -37,8 +41,11 @@ instance Exts.IsList (Maybe (List a)) where
   fromList = Just . List
   toList = undefined
 
+defaultInfo :: DependencyInfo
+defaultInfo = DependencyInfo [] anyVersion
+
 deps :: [String] -> Dependencies
-deps = Dependencies . Map.fromList . map (flip (,) AnyVersion)
+deps = Dependencies . Map.fromList . map (flip (,) defaultInfo)
 
 package :: Package
 package = Config.package "foo" "0.0.0"
@@ -59,7 +66,7 @@ withPackage content beforeAction expectation = withTempDirectory $ \dir_ -> do
   writeFile (dir </> "package.yaml") content
   withCurrentDirectory dir beforeAction
   r <- readPackageConfig (testDecodeOptions $ dir </> "package.yaml")
-  either expectationFailure (\ (DecodeResult p _ warnings) -> expectation (p, warnings)) r
+  either expectationFailure (\ (DecodeResult p _ _ warnings) -> expectation (p, warnings)) r
 
 withPackageConfig :: String -> IO () -> (Package -> Expectation) -> Expectation
 withPackageConfig content beforeAction expectation = withPackage content beforeAction (expectation . fst)
@@ -158,6 +165,38 @@ spec = do
         touch (dir </> "Setup.hs")
         getModules dir  "./." `shouldReturn` ["Foo"]
 
+  describe "toBuildTool" $ do
+    let toBuildTool_ name = runWriter $ toBuildTool "my-package" ["foo"] (name, anyVersion)
+    context "with an UnqualifiedBuildTool" $ do
+      context "when name does not match a local executable" $ do
+        it "returns a BuildTool" $ do
+          toBuildTool_ (UnqualifiedBuildTool "bar") `shouldBe` (Right (BuildTool "bar" "bar", anyVersion), [])
+
+      context "when name matches a local executable" $ do
+        it "returns a LocalBuildTool" $ do
+          toBuildTool_ (UnqualifiedBuildTool "foo") `shouldBe` (Right (LocalBuildTool "foo", anyVersion), [])
+
+      context "when name matches a legacy executable" $ do
+        it "warns" $ do
+          toBuildTool_ (UnqualifiedBuildTool "gtk2hsTypeGen") `shouldBe` (Right (BuildTool "gtk2hs-buildtools" "gtk2hsTypeGen", anyVersion), ["Usage of the unqualified build-tool name \"gtk2hsTypeGen\" is deprecated! Please use the qualified name \"gtk2hs-buildtools:gtk2hsTypeGen\" instead!"])
+
+      context "when name matches a legacy system build tool" $ do
+        it "warns" $ do
+          toBuildTool_ (UnqualifiedBuildTool "ghc") `shouldBe` (Left ("ghc", AnyVersion), ["Listing \"ghc\" under build-tools is deperecated! Please list system executables under system-build-tools instead!"])
+
+    context "with a QualifiedBuildTool" $ do
+      context "when only package matches the current package" $ do
+        it "returns a BuildTool" $ do
+          toBuildTool_ (QualifiedBuildTool "my-package" "bar") `shouldBe` (Right (BuildTool "my-package" "bar", anyVersion), [])
+
+      context "when only executable matches a local executable" $ do
+        it "returns a BuildTool" $ do
+          toBuildTool_ (QualifiedBuildTool "other-package" "foo") `shouldBe` (Right (BuildTool "other-package" "foo", anyVersion), [])
+
+      context "when both package matches the current package and executable matches a local executable" $ do
+        it "returns a LocalBuildTool" $ do
+          toBuildTool_ (QualifiedBuildTool "my-package" "foo") `shouldBe` (Right (LocalBuildTool "foo", anyVersion), [])
+
   describe "readPackageConfig" $ do
     it "warns on missing name" $ do
       withPackageWarnings_ [i|
@@ -208,7 +247,8 @@ spec = do
       withPackageConfig_ [i|
         author: John Doe
         |]
-        (`shouldBe` package {packageAuthor = ["John Doe"]})
+        (`shouldBe` package {packageAuthor = ["John Doe"]
+                            ,packageMaintainer = ["John Doe"]})
 
     it "accepts maintainer" $ do
       withPackageConfig_ [i|
@@ -401,15 +441,6 @@ spec = do
           |]
           (packageLibrary >>> (`shouldBe` Just (section library) {sectionSourceDirs = ["foo", "bar"]}))
 
-      it "accepts build-tools" $ do
-        withPackageConfig_ [i|
-          library:
-            build-tools:
-              - alex
-              - happy
-          |]
-          (packageLibrary >>> (`shouldBe` Just (section library) {sectionBuildTools = deps ["alex", "happy"]}))
-
       it "accepts default-extensions" $ do
         withPackageConfig_ [i|
           library:
@@ -437,15 +468,6 @@ spec = do
           |]
           (packageLibrary >>> (`shouldBe` Just (section library) {sectionSourceDirs = ["foo", "bar"]}))
 
-      it "accepts global build-tools" $ do
-        withPackageConfig_ [i|
-          build-tools:
-            - alex
-            - happy
-          library: {}
-          |]
-          (packageLibrary >>> (`shouldBe` Just (section library) {sectionBuildTools = deps ["alex", "happy"]}))
-
       it "allows to specify exposed" $ do
         withPackageConfig_ [i|
           library:
@@ -502,17 +524,6 @@ spec = do
           |]
           (packageExecutables >>> (`shouldBe` Map.fromList [("foo", (section (executable "Main.hs") {executableOtherModules = ["Paths_foo"]}) {sectionSourceDirs = ["foo", "bar"]})]))
 
-      it "accepts build-tools" $ do
-        withPackageConfig_ [i|
-          executables:
-            foo:
-              main: Main.hs
-              build-tools:
-                - alex
-                - happy
-          |]
-          (packageExecutables >>> (`shouldBe` Map.fromList [("foo", (section $ executable "Main.hs") {sectionBuildTools = deps ["alex", "happy"]})]))
-
       it "accepts global source-dirs" $ do
         withPackageConfig_ [i|
           source-dirs:
@@ -524,17 +535,6 @@ spec = do
           |]
           (packageExecutables >>> (`shouldBe` Map.fromList [("foo", (section (executable "Main.hs") {executableOtherModules = ["Paths_foo"]}) {sectionSourceDirs = ["foo", "bar"]})]))
 
-      it "accepts global build-tools" $ do
-        withPackageConfig_ [i|
-          build-tools:
-            - alex
-            - happy
-          executables:
-            foo:
-              main: Main.hs
-          |]
-          (packageExecutables >>> (`shouldBe` Map.fromList [("foo", (section $ executable "Main.hs") {sectionBuildTools = deps ["alex", "happy"]})]))
-
       it "accepts default-extensions" $ do
         withPackageConfig_ [i|
           executables:
